<?php

namespace Drupal\hts;

use Drupal\Component\Utility\Html;
use Drupal\Component\Utility\NestedArray;
use Drupal\Component\Utility\Xss;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\InvokeCommand;
use Drupal\Core\Ajax\ReplaceCommand;
use Drupal\Core\Form\FormStateInterface;
use Drupal\hts\Element\HierarchicalTermSelectorElement;
use Drupal\taxonomy\TermInterface;

/**
 * Defines a class for building widget form elements and settings.
 */
trait HierarchicalTermSelectorHelper {

  use TaxonomyStorageTrait;

  /**
   * Defines the default settings for this plugin.
   *
   * @return array
   *   A list of default settings, keyed by the setting name.
   */
  public static function defaultSettings(): array {
    $default_settings['hierarchical_term_selector'] = [
      'hierarchy_depth' => 2,
      'label_level' => 'label1,label2',
      'description_level' => '',
      'none_label' => HierarchicalTermSelectorElement::NONE_LABEL,
      'label_hide' => 1,
    ];
    return $default_settings;
  }

  /**
   * Returns a form to configure settings.
   *
   * @param array $form
   *   The form where the settings form is being included in.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   * @param bool $view
   *   Flag to check if comming from view.
   * @param array $settings
   *   The current plugin settings.
   *
   * @return array
   *   The form definition for the settings.
   */
  public function settingsForm(array $form, FormStateInterface $form_state, $view = FALSE, $settings = []): array {
    if(empty($settings)) {
      $default_settings = $this->defaultSettings();
      $settings = $default_settings["hierarchical_term_selector"];    
    }
    // Build hierarchical_term_selector container element.
    $element['hierarchical_term_selector'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Settings'),
      '#tree' => TRUE,
    ];

    // Inherits hierarchical_term_selector element.
    $hts_element = &$element['hierarchical_term_selector'];
    $hts_element['hierarchy_depth'] = [
      '#min' => 1,
      '#type' => 'number',
      '#title' => $this->t('Hierarchy depth'),
      '#description' => $this->t('Limits the nesting level. Use 0 to display all values. For the hierarchy like "a" -> "b" -> "c" the selection of 2 will result in "b" being the deepest option.'),
      '#default_value' => $settings['hierarchy_depth'],
      '#required' => TRUE,
    ];
    $hts_element['label_level'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Label per level'),
      '#description' => $this->t('Enter labels for each hierarchy-level separated by comma.'),
      '#default_value' => $this->t($settings['label_level']),
      '#required' => TRUE,
    ];
    $hts_element['description_level'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Description per level'),
      '#default_value' => $this->t($settings['description_level']),
      '#description' => $this->t('Enter description per hierarchy-level separated by newline.'),
    ];
    $hts_element['none_label'] = [
      '#type' => 'textarea',
      '#title' => $this->t('The "no selection" label'),
      '#default_value' => $this->t($settings['none_label']),
      '#description' => $this->t('The label for an empty option per hierarchy-level separated by newline.'),
    ];
    $hts_element['label_hide'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Hide field label'),
      '#default_value' => $settings['label_hide'] ?? FALSE,
      '#description' => $this->t('Hides the field label from the display.'),
    ];
    return $element;
  }

  /**
   * Setting per level getter.
   *
   * @param string $setting
   *   The level setting name.
   * @param string $level
   *   The level value.
   * @param array $settings
   *   The current plugin settings.
   *
   * @return string
   *   Returns the string of the settings.
   */
  public function getSettingPerLevel(string $setting, string $level, $settings = []): string {
    $value = '';
    if ($setting === 'description_level') {
      $piece = explode(PHP_EOL, $settings['description_level']);
      $value = isset($piece[$level]) ? trim($piece[$level]) : '';
    }
    elseif ($setting === 'none_label') {
      if (isset($settings['none_label'])) {
        $piece = explode(PHP_EOL, $settings['none_label']);
      }
      $value = isset($piece[$level]) ? trim($piece[$level]) : $this->t(HierarchicalTermSelectorElement::NONE_LABEL);
    }
    elseif ($setting === 'label_level') {
      $piece = explode(',', $settings['label_level']);
      $value = isset($piece[$level]) ? trim($piece[$level]) : '';
    }
    return $value;
  }

  /**
   * Returns the form for a single widget.
   *
   * @param array $element
   *   The form element.
   * @param bool $view
   *   Flag to check if comming from view.
   * @param array $settings
   *   The current plugin settings.
   * @param array $values
   *   The user input values.
   *
   * @return array
   *   The form elements for a single widget.
   */
  public function formElement(array &$element, $view = FALSE, $settings = [], $values = []): array {
    $autocomplete = NULL;
    if (isset($element['target_id'])) {
      $autocomplete = &$element['target_id'];
    }
    if (empty($settings) && method_exists($this, 'getSetting')) {
      $settings = $this->getSetting('hierarchical_term_selector');
    }

    if ($autocomplete == NULL) {
      $autocomplete = &$element;
    }
    else {
    }

    $depth = $settings['hierarchy_depth'] ?? 0;
    $delta = $autocomplete['#delta'];

    $default_value = $autocomplete['#default_value'] instanceof TermInterface ? $autocomplete['#default_value']->id() : NULL;
    $parents = $this->getParents($default_value) ? $this->getParents($default_value) : [];

    $element_id = $this->getElementIdByDelta($delta, $view);
    $autocomplete['#weight'] = -1;
    $autocomplete['#id'] = 'hts-autocomplete-' . $element_id;
    $autocomplete['#attributes']['class'][] = 'visually-hidden';

    // Build a hts element container.
    $element['hierarchical_term_selector'] = [
      '#type' => 'container',
      '#prefix' => '<div class="hts-wrapper" id="hts-' . $element_id . '">',
      '#suffix' => '</div>',
      '#attributes' => ['class' => ['hts-container']],
    ];

    // Removes the element title from the display.
    if (isset($settings['label_hide'])) {
      unset($autocomplete['#title']);
    }

    // Build a hts elements storage.
    $element['hierarchical_term_selector']['elements'] = [];
    $last_parent = $parents ? end($parents) : NULL;
    $hts_element = &$element['hierarchical_term_selector']['elements'];

    // Build initial form elements for each level.
    for ($level = 0; $level < $depth; $level++) {
      $hts_element[$level] = $this->selectElement((string) $level, $delta, $parents ?: [], $default_value, $settings, $view);
    }

    // Find last last element with parent to pre-populate options.
    foreach ($hts_element as $level => $select) {
      $current_parent = $parents[$level] ?? NULL;
      if (($current_parent == $last_parent || $last_parent == '_none')
      && isset($hts_element[$level + 1])
      && ($view == FALSE
      || (isset($values[$level])))) {
        if (!@$values[$level] != "_none") {
          $last_parent = $last_parent ?: $values[$level];
        }
        $hts_element[$level + 1]['#options'] = $this->getChildrenOptions($last_parent, (string) ($level + 1), $settings);
      }
    }

    // Attach hts styles.
    $element['#attached']['library'][] = 'hts/style';

    return $element;
  }

  /**
   * Generates select element per level.
   *
   * @param string $level
   *   The element hierarchy level.
   * @param string $delta
   *   The element delta.
   * @param array $parents
   *   The taxonomy term parent IDs.
   * @param string|null $default_value
   *   The default value of the taxonomy term ID.
   * @param array $settings
   *   The current plugin settings.
   * @param bool $view
   *   Flag to check if comming from view.
   *
   * @return array
   *   Returns the select element per level.
   */
  public function selectElement(string $level, string $delta, array $parents = [], string $default_value = NULL, $settings = [], $view = FALSE): array {
    $current_parent = $parents[$level] ?? NULL;

    $options = $this->getOptions($level, $current_parent, $view, $settings);
    $element_id = $this->getElementIdByLevel($level, $delta, $view);

    // Load term parents.
    $parents = $this->getTermStorage()->loadParents($current_parent);

    // Replace options if parents exists.
    if ($parents) {
      $parent_new = reset($parents);
      $options = $this->getChildrenOptions($parent_new->id(), $level);
    }

    $element = [
      '#type' => 'select',
      '#title' => $this->getElementLabel($level, $settings),
      '#options' => $options,
      '#empty_option' => $this->getElementNoneLabel($level, $settings),
      '#description' => $this->getElementDescription($level, $settings),
      '#prefix' => '<div class="hts-select" id="' . $element_id . '">',
      '#suffix' => '</div>',
      '#empty_value' => HierarchicalTermSelectorElement::NONE_VALUE,
      '#default_value' => $current_parent,
      '#wrapper_attributes' => ['class' => ['hts-select']],
      '#hts_id' => $element_id,
      '#hts_level' => $level,
      '#hts_delta' => $delta,
      // Fix an illegal choice issue.
      '#validated' => TRUE,
      // '#multiple' => TRUE,
    ];

    if (!$view) {
      $element['#ajax'] = [
        'callback' => [$this, 'selectElementOptionsCallback'],
        'wrapper' => $element_id,
        'progress' => [
          'type' => 'throbber',
          'message' => $this->t('Loading...'),
        ],
      ];
    }

    return $element;
  }

  /**
   * Returns the list of options.
   *
   * @param string $level
   *   The hierarchy level.
   * @param int|string $current_level
   *   The current level.
   * @param bool $view
   *   Flag to check if comming from view.
   * @param array $settings
   *   The current plugin settings.
   *
   * @return array
   *   Returns the widget options.
   */
  private function getOptions(string $level, $current_level, $view = FALSE, $settings = []): array {
    $options = [];
    $handler_settings = $settings;
    if ($view == FALSE) {
      if (property_exists($this, 'fieldDefinition')) {
        $handler_settings = $this->fieldDefinition->getSetting('handler_settings');
      }
    }

    if (!isset($handler_settings['target_bundles'])) {
      return $options;
    }

    $vocabulary_storage = $this->getVocabularyStorage();
    $parent_options = $vocabulary_storage->getToplevelTids($handler_settings['target_bundles']);

    if ($level === "0") {
      foreach ($parent_options as $parent_tid) {
        $term = $this->getTerm($parent_tid);
        $options[$parent_tid]['name'] = $term->name->value;
        $options[$parent_tid]['weight'] = $term->weight->value;
      }
    }
    elseif ($view) {
      $options = $this->getTermsByLevel((int) $level, $handler_settings['target_bundles']);
    }

    return $this->sortOptionsAlphabetically($options);
  }

  /**
   * Get the parent by taxonomy term ID.
   *
   * @param int $tid
   *   The parent taxonomy term ID.
   *
   * @return array
   *   The sorted array options.
   */
  public function getParents($tid) {
    $parents = $this->getTermStorage()->loadAllParents($tid);
    if (!$parents) {
      return [$tid];
    }

    $parent_ids = [];
    foreach ($parents as $parent) {
      $parent_ids[] = $parent->id();
    }

    // Get the array keys.
    $option_keys = array_keys($parent_ids);

    // Get the array values.
    $option_values = array_values($parent_ids);

    // Filter array with elements in reverse order.
    $option_reverse = array_reverse($option_values);

    // Combine array keys and values.
    $options = array_combine($option_keys, $option_reverse);

    return $options;
  }

  /**
   * Load the term children by parent ID.
   *
   * @param int $parent_tid
   *   The parent taxonomy term ID.
   * @param string $level
   *   The hierarchy level.
   * @param array $settings
   *   The current plugin settings.
   *
   * @return array
   *   The sorted array options.
   */
  public function getChildrenOptions($parent_tid, string $level = '', $settings = []): array {
    if ($parent_tid == "_none") {
      return $this->sortOptionsAlphabetically($this->getTermsByLevel((int) $level, $settings['target_bundles']));
    }
    $children = $this->getTermStorage()->loadChildren($parent_tid);
    $none = HierarchicalTermSelectorElement::NONE_VALUE;

    // Set initial options.
    $options = [];
    $options[$none]['name'] = $this->getElementNoneLabel($level);
    $options[$none]['weight'] = -1000;

    if ($children) {
      foreach ($children as $child) {
        $tid = $child->id();
        $options[$tid]['name'] = $child->label();
        $options[$tid]['weight'] = $child->getWeight();
      }
    }

    return $this->sortOptionsAlphabetically($options);
  }

  /**
   * Sort array in alphabetical order.
   *
   * @param array $options
   *   The select options.
   *
   * @return array
   *   The sorted array options.
   */
  public function sortOptionsAlphabetically(array $options) {
    asort($options);
    $options = array_filter(array_combine(array_keys($options), array_column($options, 'name')));
    return $options;
  }

  /**
   * An options callback to update the selected options.
   *
   * @return \Drupal\Core\Ajax\AjaxResponse
   *   The ajax response.
   */
  public function selectElementOptionsCallback(array &$form, FormStateInterface $form_state, $view = FALSE) {
    $triggering_element = $form_state->getTriggeringElement();
    $response = new AjaxResponse();

    // Prevent messages to be shown after reloading the edit form,
    // all recent messages have to be deleted.
    \Drupal::messenger()->deleteAll();

    // Get the triggering element.
    $triggering_element = $form_state->getTriggeringElement();

    // Copy the original parents.
    // Remove the last level element as well.
    $parents = $triggering_element['#array_parents'];
    array_pop($parents);

    $selected_level = $triggering_element['#hts_level'];
    $delta = $triggering_element['#hts_delta'];

    // Get the information as possible.
    if (method_exists($this, 'getSetting')) {
      $settings = $this->getSetting('hierarchical_term_selector');
    }
    $depth = $settings['hierarchy_depth'] ?? 0;
    $selected_tid = $triggering_element['#value'];
    $last_selected_tid = [];
    $none = $this->getElementNoneLabel($selected_level);

    // Pre-populates the next element options.
    for ($idx = 0; $idx < $depth; $idx++) {
      $current_level = $idx;
      $element_parents = [];
      $element_parents = $parents;

      // Update and get the next nested element array.
      if ($selected_level < $current_level) {
        array_push($element_parents, $current_level);
        $next = NestedArray::getValue($form, $element_parents);
        $next['#options'] = $this->getChildrenOptions($selected_tid, (string) $current_level);
        $next['#value'] = NULL;
        $next['#attributes']['disabled'] = $selected_level == 0 && $current_level > 1;
        $response->addCommand(new ReplaceCommand('#' . $this->getElementIdByLevel((string) $current_level, $delta), $next));
      }

      // Update and get the previous nested element array.
      if ($selected_level <= $current_level) {
        array_push($element_parents, $current_level - 1);
        $prev = NestedArray::getValue($form, $element_parents);
        if ($prev && $selected_tid == $none) {
          $last_selected_tid[] = $prev['#value'];
        }
      }
    }

    if ($last_selected_tid) {
      $last_selected_tid = array_filter($last_selected_tid);
    }

    // Get the selected taxonomy term for autocomplete field.
    $tid = $last_selected_tid[0] ?? $selected_tid;
    $term = $this->getTerm($tid);
    $selected = NULL;

    // Overrides selected if term is valid term object.
    if ($term != NULL && $term instanceof TermInterface) {
      $selected = $term->label() . ' (' . $term->id() . ')';
    }

    // Replace the autocomplete value on option selection.
    $response->addCommand(new InvokeCommand('input#hts-autocomplete-' . $this->getElementIdByDelta($delta), 'val', [$selected]));

    return $response;
  }

  /**
   * Generates ID for each element by delta.
   *
   * @param string $delta
   *   The element delta.
   * @param bool $view
   *   Flag to check if comming from view.
   *
   * @return string
   *   Returns the generated element ID.
   */
  public function getElementIdByDelta(string $delta, $view = FALSE) {
    $field_definition = NULL;
    if ($view == FALSE) {
      if (property_exists($this, 'fieldDefinition')) {
        $field_definition = $this->fieldDefinition;
      }
    }

    if ($field_definition && method_exists($field_definition, 'id')) {
      $id = str_replace('.', '_', $field_definition->id());
      $id .= '-value-' . $delta;
    }
    else {
      $id = $this->configuration['field_name'] . "-value-" . $delta;
    }
    return Html::getId($id);
  }

  /**
   * Generates ID for each element by level.
   *
   * @param string $level
   *   The element hierarchy level.
   * @param string $delta
   *   The element delta.
   * @param bool $view
   *   Flag to check if comming from view.
   *
   * @return string
   *   Returns the generated element ID.
   */
  public function getElementIdByLevel(string $level, string $delta, $view = FALSE) {
    $field_definition = NULL;
    if ($view == FALSE) {
      if (property_exists($this, 'fieldDefinition')) {
        $field_definition = $this->fieldDefinition;
      }
    }

    if ($field_definition && method_exists($field_definition, 'id')) {
      $field_id = str_replace('.', '_', $field_definition->id());
      $field_id .= '__value-' . $delta . '__level-' . $level;
    }
    else {
      $field_id = $this->configuration['field_name'] . '__value-' . $delta . '__level-' . $level;
    }
    $element_id = 'hts-select-element__' . $field_id;
    return Html::getId($element_id);
  }

  /**
   * Get the element label by level.
   *
   * @param string $level
   *   The element hierarchy level.
   * @param array $settings
   *   The current plugin settings.
   *
   * @return string|null
   *   Returns the element label.
   */
  public function getElementLabel(string $level, $settings = []) {
    $value = $this->getSettingPerLevel('label_level', $level, $settings);
    if ($value) {
      $filtered_string = Xss::filter($value);
      return $filtered_string !== '' ? $filtered_string : NULL;
    }
    return NULL;

  }

  /**
   * Get the element description by level.
   *
   * @param string $level
   *   The element hierarchy level.
   * @param array $settings
   *   The current plugin settings.
   *
   * @return string|null
   *   Returns the description label.
   */
  public function getElementDescription(string $level, $settings = []) {
    $value = $this->getSettingPerLevel('description_level', $level, $settings);
    if ($value) {
      $filtered_string = Xss::filterAdmin($value);
      return $filtered_string !== '' ? $filtered_string : NULL;
    }
    return NULL;
  }

  /**
   * Get the element "none_label" by level.
   *
   * @param string $level
   *   The element hierarchy level.
   * @param array $settings
   *   The current plugin settings.
   *
   * @return string|null
   *   Returns the "none_label" label.
   */
  public function getElementNoneLabel(string $level, $settings = []) {
    $value = $this->getSettingPerLevel('none_label', $level, $settings);
    if ($value) {
      $filtered_string = Xss::filter($value);
      return $filtered_string !== '' ? $filtered_string : NULL;
    }
    return NULL;
  }

}
