<?php

namespace Drupal\hts\Plugin\views\filter;

use Drupal\Core\Entity\Element\EntityAutocomplete;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element;
use Drupal\Core\Render\Element\Checkboxes;
use Drupal\Core\Session\AccountInterface;
use Drupal\hts\HierarchicalTermSelectorHelper;
use Drupal\hts\TaxonomyStorageTrait;
use Drupal\taxonomy\Entity\Term;
use Drupal\taxonomy\Plugin\views\filter\TaxonomyIndexTid;
use Drupal\taxonomy\TermStorageInterface;
use Drupal\taxonomy\VocabularyStorageInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a filter for hierarchical term selector extending TaxonomyIndexTid.
 *
 * @ingroup views_filter_handlers
 *
 * @ViewsFilter("hts_taxonomy_index_tid")
 */
class HtsTaxonomyIndexTid extends TaxonomyIndexTid {

  use HierarchicalTermSelectorHelper {
    defaultSettings as helperDefaultSettings;
    formElement as helperFormElement;
    settingsForm as helperSettingsForm;
  }

  use TaxonomyStorageTrait;

  /**
   * Stores the exposed input for this filter.
   *
   * @var array|null
   */
  public $validatedExposedInput = NULL;

  /**
   * The vocabulary storage.
   *
   * @var \Drupal\taxonomy\VocabularyStorageInterface
   */
  protected $vocabularyStorage;

  /**
   * The term storage.
   *
   * @var \Drupal\taxonomy\TermStorageInterface
   */
  protected $termStorage;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * The entity repository.
   *
   * @var \Drupal\Core\Entity\EntityRepositoryInterface
   */
  protected $entityRepository;

  /**
   * {@inheritdoc}
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, VocabularyStorageInterface $vocabulary_storage, TermStorageInterface $term_storage, AccountInterface $current_user = NULL, EntityRepositoryInterface $entity_repository) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $vocabulary_storage, $term_storage, $current_user);
    $this->entityRepository = $entity_repository;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager')->getStorage('taxonomy_vocabulary'),
      $container->get('entity_type.manager')->getStorage('taxonomy_term'),
      $container->get('current_user'),
      $container->get('entity.repository')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getValueOptions() {
    return $this->valueOptions;
  }

  /**
   * {@inheritdoc}
   */
  public function buildExtraOptionsForm(&$form, FormStateInterface $form_state) {
    parent::buildExtraOptionsForm($form, $form_state);
    // Add option for Hierarchical Term Selector.
    $form['type']['#options']['hts'] = $this->t('Hierarchical Term Selector');
    $form += $this->settingsForm($form, $form_state, TRUE, $this->options['hierarchical_term_selector']);
    $form['hierarchical_term_selector']['#states'] = [
      'visible' => [
        ':input[name="options[type]"]' => ['value' => 'hts'],
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  protected function valueForm(&$form, FormStateInterface $form_state) {
    $vocabulary = $this->vocabularyStorage->load($this->options['vid']);

    if (empty($vocabulary) && $this->options['limit']) {
      $form['markup'] = [
        '#markup' => '<div class="js-form-item form-item">' . $this->t('An invalid vocabulary is selected. Please change it in the options.') . '</div>',
      ];
      return;
    }
    // Let the parent class generate the base form.
    parent::valueForm($form, $form_state);

    if (($this->options['type'] !== 'hts') || !$form_state->get('exposed')) {
      // Stop further processing if the filter should not be rendered as exposed
      // filter or as Hierarchical Term Selector widget.
      return;
    }
    $options = [];
    /** @var \Drupal\Core\Entity\Query\QueryInterface $query */
    $query = $this->termStorage->getQuery('AND');
    $query->accessCheck(TRUE)
      // @todo Sorting on vocabulary properties -
      //   https://www.drupal.org/node/1821274.
      ->sort('weight')
      ->sort('name')
      ->addTag('taxonomy_term_access');
    if (!$this->currentUser->hasPermission('administer taxonomy')) {
      $query->condition('status', 1);
    }
    if ($this->options['limit']) {
      $query->condition('vid', $vocabulary->id());
    }
    /** @var \Drupal\taxonomy\TermInterface[] $terms */
    $terms = $this->termStorage->loadMultiple($query->execute());
    foreach ($terms as $term) {
      $options[$term->id()] = $this->entityRepository->getTranslationFromContext($term)->label();
    }      $form['value'] = [
      '#title' => $this->options['limit'] ? $this->t('Select terms from vocabulary @voc', ['@voc' => $vocabulary->label()]) : $this->t('Select terms'),
      '#type' => 'textfield',
      '#default_value' => EntityAutocomplete::getEntityLabels($terms),
    ];

    $values = [];
    $userInput = $form_state->getUserInput();
    foreach ($userInput as $key => $input) {
      if (is_int($key) && $input != "_none") {
        $validated[] = $input;
      }
    }
    if (!empty($validated)) {
      $values = $this->getTaxonomyTreeTerms(end($validated), $this->options['vid']);
    }

    $form['value'] = [
      '#type' => 'container',
      '#target_type' => 'taxonomy_term',
      '#selection_settings' => ['target_bundles' => [$vocabulary->id()]],
      '#tags' => TRUE,
      '#default_value' => $values,
      '#process_default_value' => FALSE,
    ];

    if ($form_state->get('exposed') == FALSE) {
      // Retain the helper option.
      $this->helper->buildOptionsForm($form, $form_state);

      // Show help text if not exposed to end users.
      $form['value']['#description'] = $this->t('Leave blank for all. Otherwise, the first selected term will be the default instead of "Any".');
    }
  }

  /**
   * {@inheritdoc}
   */
  public function buildExposedForm(&$form, FormStateInterface $form_state) {
    // Let the parent class generate the base form.
    parent::buildExposedForm($form, $form_state);
    $inputs = $form_state->getUserInput();
    if (($this->options['type'] !== 'hts') || !$form_state->get('exposed')) {
      // Stop further processing if the filter should not be rendered as exposed
      // filter or as Hierarchical Term Selector widget.
      return;
    }
    if (empty($this->options['exposed'])) {
      return;
    }

    // Build the exposed form, when it's based on an operator.
    if (!empty($this->options['expose']['use_operator']) && !empty($this->options['expose']['operator_id'])) {
      $operator = $this->options['expose']['operator_id'];
      $this->operatorForm($form, $form_state);

      // Limit the exposed operators if needed.
      if (!empty($this->options['expose']['operator_limit_selection']) &&
          !empty($this->options['expose']['operator_list'])) {

        $options = $this->operatorOptions();
        $operator_list = $this->options['expose']['operator_list'];
        $form['operator']['#options'] = array_intersect_key($options, $operator_list);
      }
      $form[$operator] = $form['operator'];

      $this->exposedTranslate($form[$operator], 'operator');

      unset($form['operator']);

      // When the operator and value forms are both in play, enclose them within
      // a wrapper.
      if (!empty($this->options['expose']['identifier'])) {
        $wrapper = $this->options['expose']['identifier'] . '_wrapper';
        $this->buildValueWrapper($form, $wrapper);
        $form[$operator]['#title_display'] = 'invisible';

        $form[$wrapper][$operator] = $form[$operator];
        unset($form[$operator]);
      }
    }

    // Build the form and set the value based on the identifier.
    if (!empty($this->options['expose']['identifier'])) {
      $value = $this->options['expose']['identifier'];
      $this->valueForm($form, $form_state);
      $form[$value] = $form['value'];

      if (isset($form[$value]['#title']) && !empty($form[$value]['#type']) && $form[$value]['#type'] != 'checkbox') {
        unset($form[$value]['#title']);
      }

      $this->exposedTranslate($form[$value], 'value');

      if ($value != 'value') {
        unset($form['value']);
      }

      // When the operator and value forms are both in play, enclose them within
      // a wrapper, for usability. Also wrap if the value form is comprised of
      // multiple elements.
      $wrapper = $value . '_wrapper';
      if ((!empty($this->options['expose']['use_operator']) && !empty($this->options['expose']['operator_id'])) || count(Element::children($form[$value]))) {
        $this->buildValueWrapper($form, $wrapper);
        $form[$wrapper][$value] = $form[$value];
        unset($form[$value]);
      }

      $element = [
        '#type' => 'container',
        '#attributes' => ['class' => ['hts-container']],
      ];
      $selectorSettings = [
        'target_bundles' => [$this->options['vid'] => $this->options['vid']],
        'sort' => [
          'field' => 'name',
          'direction' => 'asc',
        ],
        'field' => 'name',
        'direction' => 'asc',
        'auto_create' => FALSE,
        'auto_create_bundle' => '',
        'match_limit' => "10",
      ];

      $values = [];
      $userInput = $form_state->getUserInput();
      $validated = [];
      foreach ($userInput as $key => $input) {
        if ((is_int($key)) && $input != "_none") {
          $validated[] = $input;
        }
      }
      $validated = array_filter($validated, function ($value) {
        return !is_null($value);
      });
      if (!empty($validated)) {
        $values = $this->getTaxonomyTreeTerms(end($validated), $this->options['vid']);
      }

      $element = [
        'target_id' => [
          '#title' => $this->options['limit'] ? $this->t('Select terms from vocabulary @voc', ['@voc' => $this->options['vid']]) : $this->t('Select terms'),
          '#type' => 'entity_autocomplete',
          '#target_type' => 'taxonomy_term',
          "#selection_handler" => "default:taxonomy_term",
          '#selection_settings' => $selectorSettings,
          '#hierarchical_term_selector' => $this->options['hierarchical_term_selector'],
          '#default_value' => $values ? $values : '',
          '#value' => $values ? $values : '',
          '#delta' => 0,
        ],
      ];

      $settings = array_merge($this->options['hierarchical_term_selector'], $selectorSettings);
      $form[$value] = $this->formElement($element, TRUE, $settings, $inputs);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function validateExposed(&$form, FormStateInterface $form_state) {
    parent::validateExposed($form, $form_state);
    if (($this->options['type'] !== 'hts') || !$form_state->get('exposed')) {
      // Stop further processing if the filter should not be rendered as exposed
      // filter or as Hierarchical Term Selector widget.
      return;
    }
    $userInput = $form_state->getUserInput();
    $validated = [];
    foreach ($userInput as $key => $value) {
      if (is_int($key) && $value != "_none") {
        $validated[] = $value;
      }
    }
    $validated = array_filter($validated, function ($value) {
      return !is_null($value);
    });
    if (!empty($validated)) {
      // $this->validated_exposed_input = $validated;
      $this->value = $this->getTaxonomyTreeTerms(end($validated), $this->options['vid']);
      // $form_state->setValue($identifier, end($validated));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function acceptExposedInput($input) {
    if (empty($this->options['exposed'])) {
      return TRUE;
    }
    if (($this->options['type'] == 'hts')) {
      $this->validatedExposedInput = $this->value;
      return !empty($this->value);
    }
    if (!empty($this->options['expose']['use_operator']) && !empty($this->options['expose']['operator_id']) && isset($input[$this->options['expose']['operator_id']])) {
      $this->operator = $input[$this->options['expose']['operator_id']];
    }

    if (!empty($this->options['expose']['identifier'])) {
      if ($this->options['is_grouped']) {
        $value = $input[$this->options['group_info']['identifier']];
      }
      else {
        $value = $input[$this->options['expose']['identifier']];
      }

      // Various ways to check for the absence of non-required input.
      if (empty($this->options['expose']['required'])) {
        if (($this->operator == 'empty' || $this->operator == 'not empty') && $value === '') {
          $value = ' ';
        }

        if ($this->operator != 'empty' && $this->operator != 'not empty') {
          if ($value == 'All' || $value === 0 || $value === []) {
            return FALSE;
          }

          // If checkboxes are used to render this filter, do not include the
          // filter if no options are checked.
          if (is_array($value) && Checkboxes::detectEmptyCheckboxes($value)) {
            return FALSE;
          }
        }

        if (!empty($this->alwaysMultiple) && $value === '') {
          return FALSE;
        }
      }
      if (isset($value)) {
        $this->value = $value;
        if (empty($this->alwaysMultiple) && empty($this->options['expose']['multiple']) && !is_array($value)) {
          $this->value = [$value];
        }
      }
      else {
        return FALSE;
      }
    }

    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function buildExposeForm(&$form, FormStateInterface $form_state) {
    parent::buildExposeForm($form, $form_state);
    if ($this->options['type'] != 'hts') {
      unset($form['expose']['reduce']);
    }
    $form['error_message'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Display error message'),
      '#default_value' => !empty($this->options['error_message']),
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function adminSummary() {
    // Set up $this->valueOptions for the parent summary.
    $this->valueOptions = [];

    if ($this->value) {
      $this->value = array_filter($this->value);
      $terms = Term::loadMultiple($this->value);
      foreach ($terms as $term) {
        $this->valueOptions[$term->id()] = $this->entityRepository->getTranslationFromContext($term)->label();
      }
    }
    return parent::adminSummary();
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheContexts() {
    $contexts = parent::getCacheContexts();
    // The result potentially depends on term access and so is just cacheable
    // per user.
    // @todo See https://www.drupal.org/node/2352175.
    $contexts[] = 'user';

    return $contexts;
  }

  /**
   * {@inheritdoc}
   */
  public function calculateDependencies() {
    $dependencies = parent::calculateDependencies();

    $vocabulary = $this->vocabularyStorage->load($this->options['vid']);
    $dependencies[$vocabulary->getConfigDependencyKey()][] = $vocabulary->getConfigDependencyName();

    foreach ($this->termStorage->loadMultiple($this->options['value']) as $term) {
      $dependencies[$term->getConfigDependencyKey()][] = $term->getConfigDependencyName();
    }

    return $dependencies;
  }

  /**
   * {@inheritdoc}
   */
  public function submitExtraOptionsForm($form, FormStateInterface $form_state) {
    parent::submitExtraOptionsForm($form, $form_state);
    $values = $form_state->getValues();
    $options = $values["options"]["hierarchical_term_selector"]["hierarchical_term_selector"];
    if (isset($options)) {
      $this->options["hierarchical_term_selector"] = $options;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function query() {
    parent::query();

    if ($this->options['type'] == 'hts') {
      // Ensure table config.
      $this->ensureMyTable();

      if (!empty($this->value) && is_array($this->value)) {
        /** @var \Drupal\views\Plugin\views\query\Sql $query */
        $query = $this->query;
        // Change the main group operator to OR.
        $query->setGroupOperator('OR');
        // Modify the 'where' clause to use OR instead of AND for the new group.
        if (isset($query->where[$this->options['group']])) {
          $query->where[$this->options['group']]['type'] = 'OR';
          array_pop($query->where[$this->options['group']]['conditions']);
          foreach ($query->tables['node_field_data'] as $key => $queue) {
            if (str_contains($key, $this->table)) {
              unset($query->tables['node_field_data'][$key]);
            }
          }
        }
      }
    }
  }

}
