<?php

namespace Drupal\hts;

use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\taxonomy\TermInterface;
use Drupal\taxonomy\TermStorageInterface;
use Drupal\taxonomy\VocabularyStorageInterface;

/**
 * The container for representing the `hierarchical_term_selector` storage.
 */
trait TaxonomyStorageTrait {

  /**
   * An instance of the "entity_type.manager" service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Returns the entity storage.
   *
   * @param string $entity_type
   *   The entity type ID.
   *
   * @return \Drupal\Core\Entity\EntityStorageInterface
   *   The entity storage.
   */
  private function getStorage($entity_type): EntityStorageInterface {
    if (NULL === $this->entityTypeManager) {
      $this->entityTypeManager = \Drupal::service('entity_type.manager');
    }

    return $this->entityTypeManager->getStorage($entity_type);
  }

  /**
   * Returns storage of the "taxonomy_vocabulary" entities.
   *
   * @return \Drupal\taxonomy\VocabularyStorageInterface
   *   The storage of the "taxonomy_vocabulary" entities.
   */
  protected function getVocabularyStorage(): VocabularyStorageInterface {
    /** @var \Drupal\taxonomy\VocabularyStorageInterface $vocabularyStorage */
    $vocabularyStorage = $this->getStorage('taxonomy_vocabulary');
    return $vocabularyStorage;
  }

  /**
   * Returns storage of the "taxonomy_term" entities.
   *
   * @return \Drupal\taxonomy\TermStorageInterface
   *   The storage of the "taxonomy_term" entities.
   */
  protected function getTermStorage(): TermStorageInterface {
    /** @var \Drupal\taxonomy\TermStorageInterface $termStorage */
    $termStorage = $this->getStorage('taxonomy_term');
    return $termStorage;
  }

  /**
   * Returns storage of the "taxonomy_term" entities.
   *
   * @return \Drupal\taxonomy\TermInterface
   *   The storage of the "taxonomy_term" entities.
   */
  protected function getTerm(int $tid): TermInterface {
    /** @var  \Drupal\taxonomy\TermInterface $term*/
    $term = $this->getStorage('taxonomy_term')->load($tid);
    return $term;
  }

  /**
   * Get terms by hierarchy level.
   *
   * @param int $level
   *   The desired hierarchy level.
   * @param array $bundles
   *   The bundles to search.
   *
   * @return array
   *   An array of term entities at the specified level.
   */
  protected function getTermsByLevel(int $level, array $bundles): array {
    $options = [];
    foreach ($bundles as $bundle) {
      $terms = $this->getTermStorage()->loadTree($bundle, 0, $level + 1, TRUE);
      foreach ($terms as $term) {
        if ($this->getTermDepth($term) == $level) {
          $id = $term->id();
          $options[$id]['name'] = $term->label();
          $options[$id]['weight'] = $term->get('weight')->value;

        }
      }
    }
    return $options;
  }

  /**
   * Get the depth of a term in the hierarchy.
   *
   * @param \Drupal\taxonomy\TermInterface $term
   *   The term entity.
   *
   * @return int
   *   The depth of the term.
   */
  protected function getTermDepth(TermInterface $term): int {
    $depth = 0;
    while ($term->parent->target_id) {
      $term = $this->getTerm((int) $term->parent->target_id);
      $depth++;
    }

    return $depth;
  }

  /**
   * Recursively gets all the term IDs under another in a specific vocabulary.
   *
   * @param int $tid
   *   The taxonomy term ID from which to start.
   * @param string $vocabulary_id
   *   The vocabulary machine name.
   * @param \Drupal\taxonomy\TermStorageInterface|null $term_storage
   *   (optional) The term storage service. If not provided, it will be loaded.
   *
   * @return array
   *   An array of children term IDs including the given term ID.
   */
  public function getTaxonomyTreeTerms($tid, $vocabulary_id, TermStorageInterface $term_storage = NULL) {
    if (!$term_storage) {
      $term_storage = \Drupal::entityTypeManager()->getStorage('taxonomy_term');
    }

    $term_ids = [];
    $term = $term_storage->load($tid);

    if ($term && $term->bundle() === $vocabulary_id) {
      // Start with the parent term ID.
      $term_ids = [$tid];

      // Get all child terms recursively.
      $child_term_ids = $this->getChildTerms($tid, $vocabulary_id, $term_storage);
      if (!empty($child_term_ids)) {
        $term_ids = array_merge($term_ids, $child_term_ids);
      }
    }

    return $term_ids;
  }

  /**
   * Helper function to recursively get child terms.
   *
   * @param int $tid
   *   The taxonomy term ID from which to start.
   * @param string $vocabulary_id
   *   The vocabulary machine name.
   * @param \Drupal\taxonomy\TermStorageInterface $term_storage
   *   The term storage service.
   *
   * @return array
   *   An array of child term IDs.
   */
  public function getChildTerms($tid, $vocabulary_id, TermStorageInterface $term_storage) {
    $child_terms = $term_storage->loadChildren($tid);

    $term_ids = [];
    foreach ($child_terms as $child_term) {
      if ($child_term->bundle() === $vocabulary_id) {
        // Add the child term ID.
        $term_ids[] = $child_term->id();

        // Recursively get the children's children.
        $grandchildren_ids = $this->getChildTerms($child_term->id(), $vocabulary_id, $term_storage);
        if (!empty($grandchildren_ids)) {
          $term_ids = array_merge($term_ids, $grandchildren_ids);
        }
      }
    }

    return $term_ids;
  }

  /**
   * Helper function to get the sibling terms of a given term.
   *
   * @param int $tid
   *   The taxonomy term ID from which to start.
   * @param string $vocabulary_id
   *   The vocabulary machine name.
   * @param \Drupal\taxonomy\TermStorageInterface $term_storage
   *   The term storage service.
   *
   * @return array
   *   An array of sibling term IDs (including the current term).
   */
  public function getBrotherTerms($tid, $vocabulary_id, TermStorageInterface $term_storage) {
    $term = $term_storage->load($tid);
    $parent_tid = $term->parent->target_id;

    // If the term has a parent, load its siblings.
    if ($parent_tid) {
      $siblings = $term_storage->loadChildren((int) $parent_tid);
    }
    else {
      // If the term doesn't have a parent, load all terms at the top level.
      $siblings = $term_storage->loadTree($vocabulary_id, 0, 1, TRUE);
    }

    $term_ids = [];
    foreach ($siblings as $sibling) {
      if ($sibling->bundle() === $vocabulary_id && $sibling->id() != $tid) {
        $term_ids[] = $sibling->id();
      }
    }

    return $term_ids;
  }

}
